/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2005  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "pch.h"

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "Server.h"
#include "ResponseFilterChain.h"
#include "ResponseErrorInitiator.h"
#include "SharedPtr.h"
#include "ServerConnection.h"
#include "HttpRequestMetadata.h"
#include "HttpResponseMetadata.h"
#include "HttpRequestLine.h"
#include "HttpHeaderStructure.h"
#include "HttpHeaderElement.h"
#include "FilterTryList.h"
#include "CraftedResponse.h"
#include "ErrorResponse.h"
#include "ErrorDescriptor.h"
#include "ErrorCodes.h"
#include "ServerConnectionPool.h"
#include "BString.h"
#include "SplittableBuffer.h"
#include "DataChunk.h"
#include "ScopedIncDec.h"
#include "AbstractDataConsumer.h"
#include "AbstractRequestHandler.h"
#include "AbstractResponseHandler.h"
#include "HttpStreamWriter.h"
#include "HttpMessageShaper.h"
#include "GlobalState.h"
#include "ServiceContext.h"
#include "InetAddr.h"
#include "ProxyDescriptor.h"
#include "DownloadProgress.h"
#include "URI.h"
#include "StringUtils.h"
#include "InsensitiveEqual.h"
#include "NetworkActivityReporter.h"
#include "RequestLog.h"
#include "RequestTag.h"
#include "Debug.h"
#include <ace/Time_Value.h>
#include <ace/INET_Addr.h>
#include <algorithm>
#include <list>
#include <cassert>

using namespace std;

class Server::Request :
	public AbstractRequestHandler,
	public AbstractDataConsumer,
	public ResponseErrorInitiator
{
	DECLARE_NON_COPYABLE(Request)
public:
	Request(ServiceContext& context, ServerWeakPtr const& server_ptr,
		ConstRequestPtr const& request_metadata,
		RequestTag const& request_tag,
		IntrusivePtr<AbstractResponseHandler> const& response_handler,
		std::auto_ptr<FilterTryList> filter_try_list);
	
	virtual ~Request();
	
	AbstractResponseHandler& responseHandler() { return *m_ptrFilterChain; }
	
	bool hasData() const { return !m_data.empty() && !m_ptrErrDesc.get(); }
	
	SplittableBuffer& data() { return m_data; }
	
	SplittableBuffer const& data() const { return m_data; }
	
	bool isEOF() const { return m_isEOF; }
	
	bool isRestartable() const;
	
	bool isConsumingStarted() const { return m_isConsumingStarted; }
	
	void setConsumingStarted(bool val = true) { m_isConsumingStarted = val; }
	
	bool isFullyConsumed() const { return data().empty() && isEOF(); }
	
	bool restart();
	
	void setErrorResponse(IntrusivePtr<ErrorDescriptor> const& edesc) { m_ptrErrDesc = edesc; }
	
	IntrusivePtr<ErrorDescriptor> const& errorDescriptor() { return m_ptrErrDesc; }
	
	std::string const& getErrorMsg() const { return m_errorMsg; }
	
	int getClientRequestId() const { return m_ptrOrigMetadata->getClientRequestId(); }
	
	ConstRequestPtr const& getOrigMetadata() const { return m_ptrOrigMetadata; }
	
	HttpRequestLine const& getRequestLine() const { return m_ptrOrigMetadata->requestLine(); }
	
	URI const& getRequestURI() const { return m_ptrOrigMetadata->requestLine().getURI(); }
	
	HttpHeadersCollection const& getRequestHeaders() const { return m_ptrOrigMetadata->headers(); }
	
	HttpRequestMetadata const& getOutgoingMetadata() const { return *m_ptrOutgoingMetadata; }
	
	RequestTag const& getTag() const { return m_tag; }
	
	InetAddr const& getNextHopAddr() const { return m_nextHopAddr; }
	
	void updateRoute();
protected:
	virtual void raiseError(IntrusivePtr<ErrorDescriptor> const& edesc);
	
	virtual bool isErrorRaised() const;
private:
	virtual void appendBodyData(SplittableBuffer& data, bool eof);
	
	virtual void cancelRequest(IntrusivePtr<ErrorDescriptor> const& edesc);
	
	virtual void processNewData(SplittableBuffer& data, bool eof);
	
	void maybeReportClientIP(Config const& conf);
	
	void updateRoute(Config const& conf);
	
	ServiceContext& m_rContext;
	ServerWeakPtr m_ptrServer;
	ConstRequestPtr m_ptrOrigMetadata;
	std::auto_ptr<HttpRequestMetadata> m_ptrOutgoingMetadata;
	RequestTag m_tag;
	std::auto_ptr<ResponseFilterChain> m_ptrFilterChain;
	InetAddr m_nextHopAddr;
	bool m_willGoThroughProxy;
	IntrusivePtr<ErrorDescriptor> m_ptrErrDesc;
	std::string m_errorMsg;
	SplittableBuffer m_data;
	bool m_isEOF;
	bool m_isConsumingStarted;
	HttpStreamWriter m_streamWriter;
};


class Server::PhantomRequest :
	public AbstractRequestHandler,
	private NonCopyable
{
public:
	PhantomRequest(
		ConstRequestPtr const& response_metadata,
		RequestTag const& request_tag,
		IntrusivePtr<AbstractResponseHandler> const& response_handler,
		CraftedResponseGenerator const& response_generator);
	
	~PhantomRequest();
	
	AbstractResponseHandler& responseHandler() { return *m_ptrResponseHandler; }
	
	CraftedResponseGenerator& responseGenerator() { return m_responseGenerator; }
	
	URI const& getRequestURI() const { return m_ptrOrigMetadata->requestLine().getURI(); }
	
	int getClientRequestId() const { return m_ptrOrigMetadata->getClientRequestId(); }
	
	RequestTag const& getTag() const { return m_tag; }
private:
	virtual void appendBodyData(SplittableBuffer& data, bool eof);
	
	virtual void cancelRequest(IntrusivePtr<ErrorDescriptor> const& edesc);
	
	ConstRequestPtr m_ptrOrigMetadata;
	RequestTag m_tag;
	IntrusivePtr<AbstractResponseHandler> m_ptrResponseHandler;
	CraftedResponseGenerator m_responseGenerator;
};


class Server::ScopedSuspender
{
public:
	ScopedSuspender(Server& owner) : m_rOwner(owner) {
		m_rOwner.suspend();
	}
	
	~ScopedSuspender() {
		m_rOwner.resume();
	}
private:
	Server& m_rOwner;
};


class Server::RequestCanceller
{
public:
	void operator()(IntrusivePtr<Request> const& req) {
		RequestLog::logRequestCancel(req->getTag());
	}
	
	void operator()(IntrusivePtr<PhantomRequest> const& req) {
		if (req) {
			RequestLog::logRequestCancel(req->getTag());
		}
	}
};


Server::Server(ServiceContext& context, ServerTimeouts const& timeouts)
:	m_rContext(context),
	m_timeouts(timeouts),
	m_pipelineSize(0),
	m_lastResponseVersion(0, 0),
	m_readTimer(
		context.reactor(),
		sigc::mem_fun(*this, &Server::onReadTimeout),
		ACE_Time_Value(m_timeouts.getReadTimeout())
	),
	m_writeTimer(
		context.reactor(),
		sigc::mem_fun(*this, &Server::onWriteTimeout),
		ACE_Time_Value(m_timeouts.getWriteTimeout())
	),
	m_delayedActionTimer(
		context.reactor(),
		sigc::mem_fun(*this, &Server::onDelayedAction),
		ACE_Time_Value::zero
	),
	m_eventMask(Reactor::NO_EVENTS),
	m_flags(0),
	m_isWaitingForWrite(0),
	m_ptrSelf(this)
{
	m_rContext.reactor().beforeSleepSignal().connect(sigc::mem_fun(*this, &Server::updateState));
}

Server::~Server()
{
	dropConnection();
}

IntrusivePtr<AbstractRequestHandler>
Server::submitRequest(
	ConstRequestPtr const& request_metadata,
	RequestTag const& request_tag,
	IntrusivePtr<AbstractResponseHandler> const& handler)
{
	std::auto_ptr<FilterTryList> filter_try_list(new FilterTryList);
	IntrusivePtr<Request> request(new Request(
		m_rContext, m_ptrSelf, request_metadata,
		request_tag, handler, filter_try_list
	));
	inspectRequest(*request);
	enqueueRequest(request);
	updatePipeline();
	return request;
}

IntrusivePtr<AbstractRequestHandler>
Server::submitRequest(
	ConstRequestPtr const& request_metadata,
	RequestTag const& request_tag,
	IntrusivePtr<AbstractResponseHandler> const& handler,
	std::auto_ptr<FilterTryList> filter_try_list)
{
	IntrusivePtr<Request> request(new Request(
		m_rContext, m_ptrSelf, request_metadata,
		request_tag, handler, filter_try_list
	));
	inspectRequest(*request);
	enqueueRequest(request);
	updatePipeline();
	return request;
}

IntrusivePtr<AbstractRequestHandler>
Server::submitRequest(
	ConstRequestPtr const& request_metadata,
	RequestTag const& request_tag,
	IntrusivePtr<AbstractResponseHandler> const& handler,
	IntrusivePtr<ErrorDescriptor> const& error_response)
{
	std::auto_ptr<FilterTryList> filter_try_list(new FilterTryList);
	IntrusivePtr<Request> request(new Request(
		m_rContext, m_ptrSelf, request_metadata,
		request_tag, handler, filter_try_list
	));
	request->setErrorResponse(error_response);
	enqueueRequest(request);
	return request;
}

IntrusivePtr<AbstractRequestHandler>
Server::submitRequest(
	ConstRequestPtr const& request_metadata,
	RequestTag const& request_tag,
	IntrusivePtr<AbstractResponseHandler> const& handler,
	CraftedResponseGenerator const& response_generator)
{
	IntrusivePtr<PhantomRequest> request(new PhantomRequest(
		request_metadata, request_tag,
		handler, response_generator
	));
	enqueuePhantomRequest(request);
	return request;
}

void
Server::reset()
{
	storeConnection();
	assert(!m_ptrConnection.get());
	assert(!m_handlerId);
	assert(!m_readTimer.isActive());
	assert(!m_writeTimer.isActive());
	
	RequestCanceller rc;
	std::for_each(m_requestQueue.begin(), m_requestQueue.end(), rc);
	std::for_each(m_phantomRequestQueue.begin(), m_phantomRequestQueue.end(), rc);
		
	m_requestQueue.clear();
	m_phantomRequestQueue.clear();
	m_pipelineSize = 0;
	m_inputData.clear();
	m_lastResponseVersion.set(0, 0);
	m_delayedActionTimer.deactivate();
	m_flags &= ~WRITE_BUFFER_FULL;
}

void
Server::enqueueRequest(IntrusivePtr<Request> const& req)
{
	m_requestQueue.push_back(req);
	m_phantomRequestQueue.push_back(IntrusivePtr<PhantomRequest>());
	
	RequestLog::RequestType type = RequestLog::NORMAL_REQUEST;
	if (&m_rContext.scriptServer() == this) {
		type = RequestLog::INTERNAL_REQUEST;
	} else if (req->getTag()->flags().isSet(RequestTag::REQUEST_ANALYZE)) {
		type = RequestLog::ANALYZE_REQUEST;
	}
	RequestLog::logRequest(req->getTag(), req->getRequestURI(), type);
}

void
Server::enqueuePhantomRequest(IntrusivePtr<PhantomRequest> const& req)
{
	m_phantomRequestQueue.push_back(req);
	
	RequestLog::logRequest(
		req->getTag(), req->getRequestURI(),
		RequestLog::SUBST_REQUEST
	);
}

void
Server::updatePipeline()
{	
	assert(m_requestQueue.size() >= m_pipelineSize);
	
	do {
		/*
		DEBUGLOG("queue size = " << m_requestQueue.size()
			<< ", phantom queue size = " << m_phantomRequestQueue.size()
			<< ", pipeline size = " << m_pipelineSize);
		*/
		if (m_requestQueue.size() == m_pipelineSize) {
			//DEBUGLOG("nothing to do");
			if (m_pipelineSize == 0) {
				//DEBUGLOG("storing connection");
				storeConnection();
			}
			break;
		}
		
		Request& next_req = *m_requestQueue[m_pipelineSize];
		
		if (next_req.errorDescriptor().get()) {
			// This doesn't mean there won't be error requests in pipeline.
			// They may appear there as a result of cancelRequest().
			break;
		}
		
		if (m_pipelineSize > 0) {
			Request const& prev_req = *m_requestQueue[m_pipelineSize - 1];
			
			if (!m_ptrConnection.get() ||
			    m_ptrConnection->getStatus() != AsyncConnection::ESTABLISHED) {
				//DEBUGLOG("We don't have a connection yet");
				break;
			}
			if (m_ptrConnection->getNumResponses() == 0) {
				//DEBUGLOG("No responses received on this connection");
				break;
			}
			if (m_ptrConnection->getMaxResponseVersion() < HttpVersion::HTTP_1_1) {
				//DEBUGLOG("Pipelining is only useful for servers that support HTTP/1.1");
				break;
			}
			if (!m_ptrConnection->isSameAddress(next_req.getNextHopAddr())) {
				//DEBUGLOG("Request to a different address");
				break;
			}
			if (!m_ptrConnection->isExpectedToPersist()) {
				//DEBUGLOG("Connection is not expected to persist");
				break;
			}
			if (m_pipelineSize >= MAX_PIPELINE_SIZE) {
				//DEBUGLOG("MAX_PIPELINE_SIZE reached");
				break;
			}
			if (!prev_req.isFullyConsumed()) {
				//DEBUGLOG("The previous request is not complete");
				break;
			}
			if (!prev_req.isRestartable() || !next_req.isRestartable()) {
				// only requests without a body may be pipelined
				//DEBUGLOG("Either this or the previous request is not restartable");
				break;
			}
		}
		
		//DEBUGLOG("extending pipeline");
		
		++m_pipelineSize;
		
		if (m_pipelineSize == 1 && isNewConnectionRequiredFor(next_req)) {
			//DEBUGLOG("new connection is required");
			storeConnection();
		}
	} while (false);
}

void
Server::inspectRequest(Request& req)
{
	URI const& uri = req.getRequestURI();
	
	do { // check if the protocol is supported
		static char const http[] = {'h','t','t','p'};
		BString const& scheme = uri.getScheme();
		if (!StringUtils::ciEqual(scheme.begin(), scheme.end(), http, http + sizeof(http))) {
			req.setErrorResponse(IntrusivePtr<ErrorDescriptor>(
				new ErrorDescriptor(
					ErrorCodes::UNSUPPORTED_PROTOCOL,
					"unsupported protocol",
					sigc::bind(
						sigc::ptr_fun(&ErrorResponse::errUnsupportedProtocol),
						uri.toString(), uri.getScheme().toStdString()
					)
				)
			));
			return;
		}
	} while (false);
	
	do { // check for the unsupported expectation
		HttpHeadersCollection const& headers = req.getRequestHeaders();
		HttpHeader const* expect_hdr = headers.getHeaderPtr(BString("Expect"));
		if (expect_hdr) {
			HttpHeaderStructure structure(*expect_hdr);
			list<HttpHeaderElement>::iterator it = structure.elements().begin();
			list<HttpHeaderElement>::iterator const end = structure.elements().end();
			for (; it != end; ++it) {
				if (InsensitiveEqual()(it->getName(), BString("100-continue"))) {
					req.setErrorResponse(IntrusivePtr<ErrorDescriptor>(
						new ErrorDescriptor(
							ErrorCodes::EXPECTATION_FAILED,
							"expectation failed",
							sigc::bind(
								sigc::ptr_fun(&ErrorResponse::errExpectationFailed),
								uri.toString(), it->getName().toStdString()
							)
						)
					));
					return;
				}
			}
		}
	} while (false);
}

bool
Server::isNewConnectionRequiredFor(Request const& req) const
{
	if (!m_ptrConnection.get()) {
		return true;
	}
	if (!m_ptrConnection->isExpectedToPersist()) {
		return true;
	}
	if (!m_ptrConnection->isSameAddress(req.getNextHopAddr())) {
		return true;
	}
	if (!req.isRestartable()) {
		return true;
	}
	return false;
}

void
Server::requestConnectionFor(Request const& req)
{
	assert(!m_ptrConnection.get());
	if (req.isRestartable()) {
		// we always use a fresh connection for unrestartable requests
		// (requests with bodies)
		m_ptrConnection = ServerConnectionPool::instance()
			->retrieveConnection(req.getNextHopAddr());
		if (m_ptrConnection.get()) {
			assert(m_ptrConnection->getStatus() == AsyncConnection::ESTABLISHED);
			//DEBUGLOG("Reusing an existing server connection");
			connectionEstablished();
			return;
		}
	}
	//DEBUGLOG("Initiating a new server connection");
	ACE_Time_Value timeout(m_timeouts.getConnectTimeout());
	m_ptrConnection.reset(new ServerConnection);
	m_ptrConnection->setListener(this);
	m_ptrConnection->initiate(m_rContext.reactor(), req.getNextHopAddr(), &timeout);
}

bool
Server::isConnectionReady() const
{
	return m_ptrConnection.get() &&
		m_ptrConnection->getStatus() == AsyncConnection::ESTABLISHED;
}

bool
Server::isOutputReady() const
{
	return m_pipelineSize > 0 && m_requestQueue[m_pipelineSize - 1]->hasData();
}

size_t
Server::getPendingOutputSize() const
{
	size_t size = 0;
	typedef deque<IntrusivePtr<Request> >::const_iterator Iter;
	Iter it = m_requestQueue.begin();
	Iter const end = m_requestQueue.end();
	for (; it != end; ++it) {
		size += (*it)->data().size();
	}
	return size;
}

void
Server::onNewOutput()
{
	if (getPendingOutputSize() > MAX_WRITE_BUF_SIZE) {
		ScopedIncDec<int> flag_manager(m_isWaitingForWrite);
		while (isOutputReady()) {
			if (m_rContext.reactor().handleEvents() != Reactor::SUCCESS) {
				break;
			}
		}
	}
}

void
Server::processOutputData()
{
	while (m_pipelineSize > 0 && isConnectionReady()) {
		if (writeRequestData(*m_requestQueue[m_pipelineSize - 1]) == 0) {
			break;
		}
	}
}

size_t
Server::writeRequestData(Request& req)
{
	if (!req.hasData()) {
		return 0;
	}

#if defined(ACE_WIN32)
	/*
	There are certain buggy software firewalls that have problems with
	multi-fragment writes. We workaround it by coalescing fragments and
	doing a single write. Fortunately, we only have to do it for requests
	(where it has almost no effect on performance), as I haven't
	encountered a firewall that has problems with fragmented responses.
	*/
	BString buf(req.data().toBString());
	ssize_t sent = m_ptrConnection->peer().send(buf.begin(), buf.size());
	buf.clear(); // just a small optimization of memory usage
#else
	iovec iov[ACE_IOV_MAX];
	int iov_size = 0;
	for (SplittableBuffer::SpanIterator iter(req.data().spanBegin());
	     iov_size < int(sizeof(iov)/sizeof(*iov)) && !iter.isAtRightBorder();
	     ++iov_size, ++iter) {
		iov[iov_size].iov_base = (char*)iter->begin();
		iov[iov_size].iov_len = iter->size();
	}
	ssize_t sent = m_ptrConnection->peer().sendv(
		iov, iov_size
		// the timeout is infinite, but the socket is in nonblocking mode
	);
#endif
	if (sent < 0 && errno == EWOULDBLOCK) {
		m_flags |= WRITE_BUFFER_FULL;
	} else if (sent > 0) {
		m_flags &= ~WRITE_BUFFER_FULL;
		m_writeTimer.postpone();
		NetworkActivityReporter::instance()->reportActivity();
#ifdef DEBUG
		bool is_script_server = (&m_rContext.scriptServer() == this);
		if (!req.isConsumingStarted()) {
			AbstractDebugAgent::HttpMessageType msg_type =
				is_script_server
				? AbstractDebugAgent::SCRIPT_REQUEST
				: AbstractDebugAgent::OUTGOING_REQUEST;
			DBG_HTTP_MESSAGE_BEGIN(req.getOutgoingMetadata(), msg_type);
		}
		
		SplittableBuffer traf;
		req.data().splitFrontAppendBack(sent, traf);
		if (is_script_server) {
			DBG_TRAFFIC_TO_SCRIPT_SERVER(traf);
		} else {
			DBG_TRAFFIC_TO_SERVER(traf);
		}
#else
		req.data().trimFront(sent);
#endif
		
		req.setConsumingStarted();
		if (req.isFullyConsumed()) {
			//DEBUGLOG("Request " << req.getRequestURI() << " has been fully consumed");
#ifdef DEBUG
			AbstractDebugAgent::HttpMessageType msg_type =
				(&m_rContext.scriptServer() == this)
				? AbstractDebugAgent::SCRIPT_REQUEST
				: AbstractDebugAgent::OUTGOING_REQUEST;
			DBG_HTTP_MESSAGE_END(msg_type);
#endif
			updatePipeline();
		}
		return sent;
	} else {
		//DEBUGLOG("Error writing data to server");
		/*
		We don't drop the connection here because there may still be some pending data
		in the socket buffer, which may even complete some of the pending requests.
		There is also no need to unregister the WRITE_MASK, as sockets in error
		state don't generate write events.
		*/
	}
	return 0;
}

void
Server::processInputData(bool eof)
{
	while (true) {
		if (m_inputData.empty() && !eof) {
			break;
		}
		if (isInactiveState()) {
			if (m_pipelineSize != 0) {
				switchToResponseStart(m_requestQueue.front()->getOrigMetadata());
			} else {
				assert(m_requestQueue.empty());
				DEBUGLOG("Unexpected data (or EOF) received from server");
				// we don't need to call restartPendingRequests() because m_pipelineSize == 0
				dropConnection();
				m_inputData.clear();
				updatePipeline();
				break;
			}
		}
		// at this point the parser is guaranteed to be in the ResponseStart state
		assert(!m_requestQueue.empty());
		IntrusivePtr<Request> req = m_requestQueue.front();
		SplittableBuffer body_data;
		bool body_eof = false;
		bool const empty_input = m_inputData.empty();
		bool message_finishes = processNewData(m_inputData, eof, body_data, body_eof);
		req->getTag()->downloadProgress().received() += body_data.size();
		if (!req->errorDescriptor().get()) {
			if (message_finishes) {
				RequestLog::logResponseEnd(req->getTag());
#ifdef DEBUG
				AbstractDebugAgent::HttpMessageType msg_type =
					(&m_rContext.scriptServer() == this)
					? AbstractDebugAgent::SCRIPT_RESPONSE
					: AbstractDebugAgent::INCOMING_RESPONSE;
				DBG_HTTP_MESSAGE_END(msg_type);
#endif
			}
			
			if (!body_data.empty() || body_eof) {
				ScopedSuspender suspender(*this);
				req->responseHandler().processBodyData(body_data, body_eof);
			}
		}
		if (m_pipelineSize == 0 || m_requestQueue.front() != req) {
			// removed from the pipeline by the recursive invocation of reactor event loop
		} else if (req->errorDescriptor().get()) {
			// request has been cancelled
			processArtificialResponses();
			// This will drop the server connection if neccessary.
			// Dropping the connection also causes switchToInactive() to be called.
		} else if (message_finishes) {
			DEBUGLOG("Response finishes: " << req->getRequestURI());
			assert(isInactiveState()); // message_finishes implies that
			if (m_ptrConnection.get()) {
				m_ptrConnection->registerResponse(m_lastResponseVersion);
			}
			shiftRequestQueue();
			if (eof || !m_ptrConnection.get() || !m_ptrConnection->isExpectedToPersist()) {
				restartPendingRequests();
				dropConnection();
				// m_inputData may still contain some useful data
			}
			processArtificialResponses();
			updatePipeline();
		} else if (isErrorState()) {
			// keep in mind: an eof while in ResponseStart state generates an error
			/*
			DEBUGLOG("ERROR parsing response: "
				<< (errorDescriptor().get()
				? errorDescriptor()->getErrorMessage()
				: errorMessage()));
			DEBUGLOG("empty_input = " << empty_input
				<< ", have_connection = " << (m_ptrConnection.get() ? true : false)
				<< ", num_responses = " << (m_ptrConnection.get() ? m_ptrConnection->getNumResponses() : 0)
				<< ", request_restartable = " << m_requestQueue.front()->isRestartable());
			*/
			/*
			This could be an OK situation:
			You have two requests in queue, the server respondes on the first one
			and immediately closes the connection. An EOF while in ResponseStart state
			generates an error.
			*/
			if (empty_input && (m_ptrConnection.get()
			    ? m_ptrConnection->getNumResponses() > 0 : true)
			    && m_requestQueue.front()->isRestartable()) {
				/*
				Q: Why do we check for empty_input?
				A: The only case when an error is generated after processing an empty input
				is in the ResponseStart (or RequestStart) state.
				Q: Why do we check for getNumResponses() > 0?
				A: Because otherwise we could get an infinite loop if the server always
				closes the connection instead of answering the response.
				Q: Why do we allow the lack of connection?
				A: Because if the connection that we've dropped had zero responses,
				we'd already generated the appropriate response and moved on to the next request.
				*/
			} else {
				/*
				DEBUGLOG("Request URI: " << m_requestQueue.front()->getRequestURI());
				DEBUGLOG("empty_input = " << empty_input << ", got_responses = "
					<< (m_ptrConnection->getNumResponses() > 0));
				*/
#ifdef DEBUG
				AbstractDebugAgent::HttpMessageType msg_type =
					(&m_rContext.scriptServer() == this)
					? AbstractDebugAgent::SCRIPT_RESPONSE
					: AbstractDebugAgent::INCOMING_RESPONSE;
				DBG_HTTP_MESSAGE_ERROR(msg_type);
#endif
				if (errorDescriptor().get()) {
					answerWithError(errorDescriptor());
				} else {
					answerWithError(IntrusivePtr<ErrorDescriptor>(
						new ErrorDescriptor(
							ErrorCodes::ERROR_PARSING_RESPONSE,
							"error parsing response: "+errorMessage(),
							sigc::bind(
								sigc::ptr_fun(&ErrorResponse::errParsingResponse),
								m_requestQueue.front()->getRequestURI().toString(),
								errorMessage()
							)
						)
					));
				}
			}
			restartPendingRequests();
			dropConnection();
			m_inputData.clear();
			updatePipeline();
			break;
		}
		if (!message_finishes) {
			// This means more input data is needed.
			// Example: the input ends in the middle of a header.
			break;
		} else if (eof && m_inputData.empty()) {
			// We have to break here, or the eof will act
			// on the next response (if any).
			break;
		}
	}
	if (eof) {
		dropConnection();
	}
}

void
Server::shiftRequestQueue()
{
	assert(!m_requestQueue.empty());
	assert(!m_phantomRequestQueue.empty());
	m_requestQueue.pop_front();
	m_phantomRequestQueue.pop_front();
	if (m_pipelineSize > 0) {
		--m_pipelineSize;
	}
}

bool
Server::connectionMayBeStored() const
{
	if (!m_ptrConnection.get()) {
		return false;
	}
	if (m_ptrConnection->getStatus() != AsyncConnection::ESTABLISHED) {
		return false;
	}
	if (!m_ptrConnection->isExpectedToPersist()) {
		return false;
	}
	if (m_ptrConnection->getNumResponses() == 0) {
		return false;
	}
	if (m_pipelineSize != 0 && m_requestQueue.front()->isConsumingStarted()) {
		return false;
	}
	return true;
}

void
Server::storeConnection()
{
	if (m_ptrConnection.get()) {
		if (!connectionMayBeStored()) {
			dropConnection();
		} else {
			auto_ptr<ServerConnection> conn(retrieveConnection());
			assert(conn.get());
			ServerConnectionPool::instance()->storeConnection(conn);
			//DEBUGLOG("connection stored");
		}
	}
}

void
Server::dropConnection()
{
	retrieveConnection();
}

std::auto_ptr<ServerConnection>
Server::retrieveConnection()
{
	auto_ptr<ServerConnection> conn;
	
	if (m_ptrConnection.get()) {
		m_readTimer.deactivate();
		m_writeTimer.deactivate();
		
		if (m_handlerId) {
			m_rContext.reactor().unregisterHandler(m_handlerId);
			m_handlerId = ReactorHandlerId();
		}
		
		conn = m_ptrConnection;
		m_eventMask = Reactor::NO_EVENTS;
		resetFlags();
		switchToInactive();
	}
	
	return conn;
}

void
Server::restartPendingRequests()
{
	for (size_t i = 0; i < m_pipelineSize; ++i) {
		m_requestQueue[i]->restart();
	}
	m_pipelineSize = 0;
}

void
Server::answerWithError(IntrusivePtr<ErrorDescriptor> const& edesc)
{
	assert(!m_requestQueue.empty());
	
	Request& req = *m_requestQueue.front();
	
	if (edesc->getErrorCode() == ErrorCodes::REQUEST_CANCEL_TO_SAVE_TRAFFIC) {
		RequestLog::logResponseEnd(req.getTag());
	} else {
		RequestLog::logResponseError(req.getTag());
	}
	
	{
		ScopedSuspender suspender(*this);
		req.responseHandler().processError(edesc);
	}
	
	shiftRequestQueue();
}

bool
Server::isRequestAtPipelineStart(Request const* req) const
{
	return (m_pipelineSize > 0 && &*m_requestQueue.front() == req);
}

void
Server::onReadTimeout()
{
	DEBUGLOG("server read timeout");
	onTimeout();
}

void
Server::onWriteTimeout()
{
	DEBUGLOG("server write timeout");
	onTimeout();
}

void
Server::onTimeout()
{
	URI const& uri = m_requestQueue.front()->getRequestURI();
	answerWithError(IntrusivePtr<ErrorDescriptor>(
		new ErrorDescriptor(
			ErrorCodes::SERVER_TIMEOUT,
			"server timeout",
			sigc::bind(
				sigc::ptr_fun(&ErrorResponse::errServerTimeout),
				uri.toString()
			)
		)
	));
	restartPendingRequests();
	dropConnection();
	updatePipeline();
}

void
Server::updateState()
{
	/*
	upadateState() MUST NOT generate any output!!!
	Server::updateState() and Client::updateState() are called just before
	going to sleep to update their event masks. If one of them generates
	output for the other, that other one would need to update its event
	mask, but it may not have a chance to do so, as it may have already run
	in the current reactor dispatch.
	
	Functions that can generate output include processArtificialResponses()
	and requestConnectionFor().
	*/
	
	if (!(m_flags & READING_SUSPENDED)) {
		if (isArtificialResponseRequired()) {
			m_delayedActionTimer.ensureActive();
		} else if (m_pipelineSize > 0 && !m_ptrConnection.get()) {
			assert(m_pipelineSize == 1);
			Request& req = *m_requestQueue.front();
			if (req.hasData()) {
				m_delayedActionTimer.ensureActive();
			}
		}
	}
	
	if (!isConnectionReady()) {
		return; // events and timers are already unregistered / deactivated
	}
	
	if (!(m_flags & WRITE_BUFFER_FULL) && isOutputReady()) {
		// Necessary on the edge-triggered WFMOReactor.
		processOutputData();
	}
	
	
	Reactor::IOEvents mask = m_eventMask;
	
	if (m_flags & READING_SUSPENDED) {
		mask &= ~Reactor::READ;
	} else {
		mask |= Reactor::READ;
	}
	if (isOutputReady()) {
		mask |= Reactor::WRITE;
	} else {
		mask &= ~Reactor::WRITE;
		if (m_isWaitingForWrite) {
			m_rContext.reactor().wakeup();
		}
		
		m_flags &= ~WRITE_BUFFER_FULL;
		// It may actually be full, but we don't want to miss
		// an edge-triggered (in case of WFMOReactor) write event.
	}
	
	if ((mask & Reactor::READ) && m_pipelineSize > 0 &&
	    m_requestQueue.front()->isFullyConsumed()) {
		m_readTimer.ensureActive();
	} else {
		m_readTimer.deactivate();
	}
	if (mask & Reactor::WRITE) {
		m_writeTimer.ensureActive();
	} else {
		m_writeTimer.deactivate();
	}
	
	if (mask != m_eventMask) {
		m_rContext.reactor().setEvents(m_handlerId, mask);
		m_eventMask = mask;
	}
}

bool
Server::processLeadingPhantomResponse()
{
	if (!m_phantomRequestQueue.empty()) {
		if (PhantomRequest* req = m_phantomRequestQueue.front().get()) {
			auto_ptr<CraftedResponse> response = req->responseGenerator()();
			req->getTag()->flags().set(RequestTag::RESPONSE_CRAFTED);
			response->metadata().setClientRequestId(req->getClientRequestId());
#ifdef DEBUG
			response->metadata().headers().setHeader(
				BString("X-Request-URI"),
				req->getRequestURI().toBString()
			);
#endif
			{
				ScopedSuspender suspender(*this);
				req->responseHandler().processResponseMetadata(
					response->retrieveMetadata()
				);
				req->responseHandler().processBodyData(
					response->body(), true
				);
			}
			
			RequestLog::logResponseBegin(req->getTag(), 200);
			RequestLog::logResponseEnd(req->getTag());
			m_phantomRequestQueue.pop_front();
			return true;
		}
	}
	return false;
}

bool
Server::processLeadingErrorResponse()
{
	if (!m_requestQueue.empty()) {
		Request& req = *m_requestQueue.front();
		if (req.errorDescriptor().get()) {
			if (!req.isConsumingStarted()) {
				answerWithError(req.errorDescriptor());
			} else {
#ifdef DEBUG
				if (!req.isFullyConsumed()) {
					AbstractDebugAgent::HttpMessageType msg_type =
						(&m_rContext.scriptServer() == this)
						? AbstractDebugAgent::SCRIPT_REQUEST
						: AbstractDebugAgent::OUTGOING_REQUEST;
					DBG_HTTP_MESSAGE_ERROR(msg_type);
				}
#endif
				answerWithError(req.errorDescriptor());
				restartPendingRequests();
				dropConnection();
				m_inputData.clear();
			}
			return true;
		}
	}
	return false;
}

void
Server::processArtificialResponses()
{
	bool need_update_pipeline = false;
	while (true) {
		bool phantom_processed = processLeadingPhantomResponse();
		bool error_processed = processLeadingErrorResponse();
		if (phantom_processed || error_processed) {
			need_update_pipeline |= error_processed;
			continue;
		}
		break;
	};
	if (need_update_pipeline) {
		updatePipeline();
	}
}

bool
Server::isArtificialResponseRequired() const
{
	if (!m_phantomRequestQueue.empty() && m_phantomRequestQueue.front().get()) {
		return true;
	}
	if (!m_requestQueue.empty() && m_requestQueue.front()->errorDescriptor().get()) {
		return true;
	}
	return false;
}

void
Server::eventGotProvisionalResponse(auto_ptr<HttpResponseMetadata> metadata)
{
	Request& req = *m_requestQueue.front();
	ScopedSuspender suspender(*this);
	req.responseHandler().processProvisionalResponse(metadata);
}

void
Server::eventGotMetadata(auto_ptr<HttpResponseMetadata> metadata, bool is_persistent)
{
	m_lastResponseVersion = metadata->statusLine().getHttpVersion();
	Request& req = *m_requestQueue.front();
	
#ifdef DEBUG
	DEBUGLOG("Response begins: " << req.getRequestURI());
	
	metadata->setClientRequestId(req.getClientRequestId());
	
	AbstractDebugAgent::HttpMessageType msg_type =
		(&m_rContext.scriptServer() == this)
		? AbstractDebugAgent::SCRIPT_RESPONSE
		: AbstractDebugAgent::INCOMING_RESPONSE;
	DBG_HTTP_MESSAGE_BEGIN(*metadata, msg_type);
	
	metadata->headers().setHeader(
		BString("X-Request-URI"),
		req.getRequestURI().toBString()
	);
#endif
	
	HttpMessageShaper::shapeResponse(*metadata, req.getRequestLine());
	
	if (!is_persistent && m_ptrConnection.get()) {
		m_ptrConnection->toBeClosed();
	}
	
	DownloadProgress& dp = req.getTag()->downloadProgress();
	dp.received() = 0;
	if (metadata->getBodyStatus() == HttpResponseMetadata::BODY_SIZE_UNKNOWN) {
		dp.total() = DownloadProgress::UNDEFINED;
	} else {
		dp.total() = metadata->getBodySize();
	}
	
	RequestLog::logResponseBegin(req.getTag(), metadata->statusLine().getCode());
	
	ScopedSuspender suspender(*this);
	req.responseHandler().processResponseMetadata(metadata);
}

void
Server::connectionEstablished()
{
	//DEBUGLOG("Connection established");
	m_ptrConnection->setTcpNoDelay();
	m_ptrConnection->peer().enable(ACE_NONBLOCK);
	ACE_HANDLE handle = m_ptrConnection->peer().get_handle();
	m_flags &= ~WRITE_BUFFER_FULL;
	m_eventMask = Reactor::READ|Reactor::WRITE;
	// here we don't check if the current request has data available,
	// as we do it before initiating a connection
	try {
		m_handlerId = m_rContext.reactor().registerHandler(
			handle, IntrusivePtr<EventHandlerBase>(this),
			m_eventMask
		);
	} catch (Reactor::Exception& e) {
		string problem("Could not register a socket in the reactor: "+string(e.what()));
		DEBUGLOG2(problem);
		m_ptrConnection.reset(0);
		Request const& req = *m_requestQueue.front();
		answerWithError(IntrusivePtr<ErrorDescriptor>(
			new ErrorDescriptor(
				ErrorCodes::UNKNOWN_ERROR,
				problem,
				sigc::bind(
					sigc::ptr_fun(&ErrorResponse::errGenericError),
					req.getRequestURI().toString(), problem,
					int(HttpStatusLine::SC_INTERNAL_SERVER_ERROR)
				)
			)
		));
		return;
	}
}

void
Server::connectionFailed(FailureCode fcode, int errnum)
{
	DEBUGLOG("Connection failed");
	assert(m_pipelineSize > 0);
	assert(m_requestQueue.size() >= m_pipelineSize);
	
	m_ptrConnection.reset(0);
	Request const& req = *m_requestQueue.front();
	URI const& uri = req.getRequestURI();
	
	switch (fcode) {
		case FAIL_DNS: {
			answerWithError(IntrusivePtr<ErrorDescriptor>(
				new ErrorDescriptor(
					ErrorCodes::DNS_ERROR,
					"could not resolve host name",
					sigc::bind(
						sigc::ptr_fun(&ErrorResponse::errHostNameResolution),
						uri.toString(), req.getNextHopAddr().getHost()
					)
				)
			));
			break;
		}
		case FAIL_TIMEOUT: {
			answerWithError(IntrusivePtr<ErrorDescriptor>(
				new ErrorDescriptor(
					ErrorCodes::CONNECT_TIMEOUT,
					"connection timeout",
					sigc::bind(
						sigc::ptr_fun(&ErrorResponse::errConnectionTimeout),
						uri.toString(),
						req.getNextHopAddr().getHost(),
						req.getNextHopAddr().getPort()
					)
				)
			));
			break;
		}
		case FAIL_REFUSED:
		case FAIL_UNREACHABLE: {
			char const* syserr = (fcode == FAIL_REFUSED)
				? "connection refused"
				: "destination host unreachable";
			answerWithError(IntrusivePtr<ErrorDescriptor>(
				new ErrorDescriptor(
					ErrorCodes::CONNECT_FAILED,
					"connection failed",
					sigc::bind(
						sigc::ptr_fun(&ErrorResponse::errConnectionFailed),
						uri.toString(), req.getNextHopAddr().getHost(),
						req.getNextHopAddr().getPort(), syserr
					)
				)
			));
			break;
		}
		case FAIL_OTHER:
		default: {
			char const* syserr = ACE::sock_error(errnum);
			if (!syserr) {
				syserr = ACE_OS::strerror(errnum);
			}
			answerWithError(IntrusivePtr<ErrorDescriptor>(
				new ErrorDescriptor(
					ErrorCodes::CONNECT_FAILED,
					"connection failed",
					sigc::bind(
						sigc::ptr_fun(&ErrorResponse::errConnectionFailed),
						uri.toString(), req.getNextHopAddr().getHost(),
						req.getNextHopAddr().getPort(), syserr
					)
				)
			));
			break;
		}
	}
	updatePipeline();
}

void
Server::handleRead(ACE_HANDLE)
{
	if (m_flags & READING_SUSPENDED) {
		// this may happen because the event mask is updated after
		// all of the triggered events are processed
		return;
	}
	processArtificialResponses();
	auto_ptr<DataChunk> chunk = DataChunk::create(READ_BUF_SIZE);
	ssize_t received = m_ptrConnection->peer().recv(
		chunk->getDataAddr(), chunk->getDataSize()
		// the timeout is infinite, but the socket is in nonblocking mode
	);
	if (received > 0) {
		NetworkActivityReporter::instance()->reportActivity();
		chunk = DataChunk::resize(chunk, received);
#ifdef DEBUG
		SplittableBuffer traf;
		traf.append(chunk);
		if (this == &m_rContext.scriptServer()) {
			DBG_TRAFFIC_FROM_SCRIPT_SERVER(traf);
		} else {
			DBG_TRAFFIC_FROM_SERVER(traf);
		}
		m_inputData.appendDestructive(traf);
#else
		m_inputData.append(chunk);
#endif
	} else if (received < 0 && errno == EWOULDBLOCK) {
		// not a problem
		return;
	}
	m_readTimer.postpone();
	bool eof = (received <= 0);
	processInputData(eof);
}

void
Server::handleWrite(ACE_HANDLE)
{
	m_flags &= ~WRITE_BUFFER_FULL;
	processOutputData();
}

void
Server::onDelayedAction()
{
	if (m_flags & READING_SUSPENDED) {
		/*
		The intent of READING_SUSPENDED is to prevent the server
		from providing more data to the response handler which
		is blocked in a recursive handleEvents().
		*/
		return;
	}
	
	if (m_pipelineSize > 0 && !m_ptrConnection.get()) {
		assert(m_pipelineSize == 1);
		m_requestQueue.front()->updateRoute();
		// This may trigger a proxy configuration error, which
		// will be handled by processArtificialResponses() below.
	}
	
	if (isArtificialResponseRequired()) {
		processArtificialResponses();
	}
	
	assert(!isArtificialResponseRequired());
	
	if (m_pipelineSize > 0 && !m_ptrConnection.get()) {
		assert(m_pipelineSize == 1);
		Request& req = *m_requestQueue.front();
		if (req.hasData()) {
			requestConnectionFor(req);
		}
	}
}


/*=========================== Server::Request ==============================*/

Server::Request::Request(
	ServiceContext& context, ServerWeakPtr const& server_ptr,
	ConstRequestPtr const& request_metadata,
	RequestTag const& request_tag,
	IntrusivePtr<AbstractResponseHandler> const& response_handler,
	auto_ptr<FilterTryList> filter_try_list)
:	m_rContext(context),
	m_ptrServer(server_ptr),
	m_ptrOrigMetadata(request_metadata),
	m_ptrOutgoingMetadata(new HttpRequestMetadata(*request_metadata)),
	m_tag(request_tag),
	m_ptrFilterChain(new ResponseFilterChain(
		context, m_ptrOrigMetadata, request_tag,
		filter_try_list, response_handler, *this
	)),
	m_nextHopAddr("", -1),
	m_willGoThroughProxy(false),
	m_isEOF(false),
	m_isConsumingStarted(false),
	m_streamWriter(*this)
{
	{
		GlobalState::ReadAccessor global_state;
		maybeReportClientIP(global_state->config());
		updateRoute(global_state->config());
	}
	m_streamWriter.startRequest(*m_ptrOutgoingMetadata, false, m_willGoThroughProxy, false);
}

Server::Request::~Request()
{
}

void
Server::Request::updateRoute()
{
	updateRoute(GlobalState::ReadAccessor()->config());
}

void
Server::Request::updateRoute(Config const& conf)
{
	URI const& uri = getRequestURI();
	std::string host = uri.getHost().toStdString();
	
	auto_ptr<ProxyDescriptor> proxy;
	
	if (!conf.getNextHopProxy().hasProblem()) {
		proxy = conf.getProxyFor(host);
	} else {
		std::string problem = conf.getNextHopProxy().getProblem();
		setErrorResponse(IntrusivePtr<ErrorDescriptor>(
			new ErrorDescriptor(
				ErrorCodes::REMOTE_PROXY_CONFIG_PROBLEM,
				problem,
				sigc::bind(
					sigc::ptr_fun(&ErrorResponse::errRemoteProxyConfig),
					getRequestURI().toString(), problem
				)
			)
		));
	}
	
	if (proxy.get()) {
		m_willGoThroughProxy = true;
		m_nextHopAddr = proxy->getProxyAddr();
	} else {
		m_willGoThroughProxy = false;
		m_nextHopAddr.set(host, uri.guessPort());
	}
}

void
Server::Request::raiseError(IntrusivePtr<ErrorDescriptor> const& edesc)
{
	m_ptrErrDesc = edesc;
}

bool
Server::Request::isErrorRaised() const
{
	return m_ptrErrDesc.get();
}

void
Server::Request::appendBodyData(SplittableBuffer& data, bool eof)
{
	if (errorDescriptor().get()) {
		return;
	}
	
	m_streamWriter.appendBodyData(data, eof);
	if (m_streamWriter.getBufferedBodyDataSize() > MAX_BUFFERED_REQUEST_BODY) {
		string problem("request body too big");
		cancelRequest(IntrusivePtr<ErrorDescriptor>(
			new ErrorDescriptor(
				ErrorCodes::REQUEST_BODY_TOO_BIG, problem,
				sigc::bind(
					sigc::ptr_fun(&ErrorResponse::errParsingRequest),
					getRequestURI().toString(), problem
				)
			)
		));
		return;
	}
}

void
Server::Request::cancelRequest(IntrusivePtr<ErrorDescriptor> const& edesc)
{
	setErrorResponse(edesc);
}

void
Server::Request::processNewData(SplittableBuffer& data, bool eof)
{
	m_data.appendDestructive(data);
	m_isEOF |= eof;
	
	if (Server* server = m_ptrServer.get()) {
		server->onNewOutput();
	} else {
		// the server is being (or have already been) destroyed
		m_data.clear();
	}
}

bool
Server::Request::isRestartable() const
{
	return !m_ptrOutgoingMetadata->hasBody();
}

bool
Server::Request::restart()
{
	if (!isRestartable()) {
		return false;
	}
	m_data.clear();
	m_isConsumingStarted = false;
	updateRoute(); // user settings might have changed
	m_streamWriter.startRequest(*m_ptrOutgoingMetadata, false, m_willGoThroughProxy, false);
	SplittableBuffer body_data;
	m_streamWriter.appendBodyData(body_data, true);
	return true;
}

void
Server::Request::maybeReportClientIP(Config const& conf)
{
	Config::ReportClientIP report = conf.getReportClientIP();
	if (report == Config::REPORT_IP_FIXED) {
		m_ptrOutgoingMetadata->headers().addHeader(
			BString("X-Forwarded-For"),
			BString(conf.getFixedClientIP())
		);
	} else if (report == Config::REPORT_IP_ON) {
		enum { MAX_IP_LEN = 46 }; // 46 is the maximum length for IPV6
		auto_ptr<DataChunk> chunk = DataChunk::create(MAX_IP_LEN);
		if (m_rContext.clientAddr().get_host_addr(chunk->getDataAddr(), MAX_IP_LEN)) {
			size_t const len = strlen(chunk->getDataAddr());
			m_ptrOutgoingMetadata->headers().addHeader(
				BString("X-Forwarded-For"),
				BString(DataChunk::resize(chunk, len))
			);
		}
	}
}


/*======================== Server::PhantomRequest ============================*/

Server::PhantomRequest::PhantomRequest(
	ConstRequestPtr const& request_metadata,
	RequestTag const& request_tag,
	IntrusivePtr<AbstractResponseHandler> const& response_handler,
	CraftedResponseGenerator const& response_generator)
:	m_ptrOrigMetadata(request_metadata),
	m_tag(request_tag),
	m_ptrResponseHandler(response_handler),
	m_responseGenerator(response_generator)
{
}

Server::PhantomRequest::~PhantomRequest()
{
}

void
Server::PhantomRequest::appendBodyData(SplittableBuffer& data, bool eof)
{
}

void
Server::PhantomRequest::cancelRequest(IntrusivePtr<ErrorDescriptor> const& edesc)
{
}
